import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * Supports training a decision tree using an adapted version of the C4.5
 * algorithm as well as running new observations through the decision tree to
 * predict a label.
 * 
 * @author Hank Feild
 * @author YOU
 */
public class DecisionTree {

    /**
     * Supports observations that are both labeled and unlabeled (e.g., for 
     * unseen observations).
     */
    public class Observation {
        public ArrayList<Double> features;
        public String label;

        /**
         * Sets the features.
         * 
         * @param features A list of feature values.
         * @param label A class label for this observation. Us null for an
         *              observation without a label.
         */
        public Observation(ArrayList<Double> features, String label) {
            this.features = features;
            this.label = label;
        }

        /**
         * @return A CSV string for of this Observation. All features are
         *         listed in order followed by the label if not null.
         */
        public String toString() {
            StringBuffer output = new StringBuffer();

            for(int i = 0; i < features.size(); i++){
                output.append(features.get(i));
                if(i < features.size()-1)
                    output.append(",");
            }
            if(label != null)
                output.append(",").append(label);

            return output.toString();
        }
    }

    /**
     * Represents a data file: a header with names of columns and a list of 
     * observations.
     */
    public class Dataset {
        public ArrayList<String> columnNames;
        public ArrayList<Observation> observations;

        /**
         * Sets the data members.
         * @param columnNames A list of column names.
         * @param observations A list of observations.
         */
        public Dataset(ArrayList<String> columnNames, 
                       ArrayList<Observation> observations) {
            this.columnNames = columnNames;
            this.observations = observations;
        }

        /**
         * @return A CSV version of the header.
         */
        public String columnNamesAsCSV() {
            StringBuffer output = new StringBuffer();

            for(int i = 0; i < columnNames.size(); i++){
                output.append(columnNames.get(i));
                if(i < columnNames.size()-1)
                    output.append(",");
            }
                
            return output.toString();
        }
    }



    /**
     * Generate a list of observations from the data file.
     * 
     * @param filename The name of the file to parse. Should have a header and be in
     *                 comma separated value (CSV) format.
     * @param hasLabel If true, the last column will be used as the label for each
     *                 observation and all other columns will be features. If false,
     *                 *all* columns will be used as features.
     * @return The observations and header (column names).
     * @throws IOException
     */
    public Dataset parseDataFile(String filename, boolean hasLabel) 
        throws IOException {

        ArrayList<String> columnNames = new ArrayList<String>();
        ArrayList<Observation> observations = new ArrayList<Observation>();
        BufferedReader reader = new BufferedReader(new FileReader(filename));

        for(String col : reader.readLine().split(",")){
            columnNames.add(col);
        }

        while(reader.ready()){
            String[] columns = reader.readLine().split(",");
            ArrayList<Double> features = new ArrayList<Double>();
            String label = null;
            if(hasLabel){
                for(int i = 0; i < columns.length-1; i++){
                    features.add(Double.parseDouble(columns[i]));
                }
                label = columns[columns.length-1];
            } else {
                for(int i = 0; i < columns.length; i++){
                    features.add(Double.parseDouble(columns[i]));
                }
            }
            
            observations.add(new Observation(features, label));
        }
        reader.close();

        return new Dataset(columnNames, observations);
    }

    /**
     * Represents a node in the decision tree.
     */
    public class Node {
        HashMap<String,Integer> labelDistribution;
        int n, featureIndex;
        double threshold;
        Node lessThanEqualChild, greaterThanChild;

        /**
         * Sets the Node's data members for a leaf node.
         * 
         * @param labelDistribution The distribution over labels represented in
         *                          this subtree.
         * @param n The total number of labels represented in this subtree.
         */
        public Node(HashMap<String,Integer> labelDistribution, int n){
            this(labelDistribution, n, 0, 0, null, null);
        }

        /**
         * Sets the Node's data members for an internal node.
         * 
         * @param labelDistribution The distribution over labels represented in
         *                          this subtree.
         * @param n The total number of labels represented in this subtree.
         * @param featureIndex The index of the feature to split on (only for
         *                     internal nodes).
         * @param threshold The threshold of the feature to use to determine
         *                  which child to direct new observations to (only for
         *                  internal nodes).
         * @param lessThanChild The child to send observations to with a feature
         *                      value less than or equal to the threshold (only
         *                      for internal nodes).
         * @param greatherThanChild The child to send observations to with a
         *                          feature value greater than the threshold
         *                          (only for internal nodes).
         */
        public Node(HashMap<String,Integer> labelDistribution, int n, 
                    int featureIndex, double threshold, Node lessThanEqualChild, 
                    Node greaterThanChild){
            
            this.labelDistribution = labelDistribution;
            this.n = n;
            this.featureIndex = featureIndex;
            this.threshold = threshold;
            this.lessThanEqualChild = lessThanEqualChild;
            this.greaterThanChild = greaterThanChild;
        }

        /**
         * @return True if this is a leaf node.
         */
        public boolean isLeafNode(){
            return lessThanEqualChild == null;
        }
    }

    Node root;

    /**
     * Initializes the root of the tree to null.
     */
    public DecisionTree() {
        root = null;
    }

    /**
     * Trains a decision tree on the set of observations and stores this in
     * `self.root`.
     * 
     * @param observations A list of Observations with labels.
     */
    public void train(ArrayList<Observation> observations){
        root = build(observations);
        prune();
    }

    /**
     * Builds a decision tree on the set of observations.
     * 
     * @param observations A list of Observations with labels.
     * @return A Node representing a subtree or leaf.
     */
    public Node build(ArrayList<Observation> observations){
        return null;
    }

    /**
     * Prunes the tree rooted at `self.root`.
     */
    public void prune(){

    }

    /**
     * Finds the threshold for the given feature that does the best job 
     * splits observations into two sets according to the `evaluate` function. 
        
     * @param observations A list of Observation instances, with labels.
     * @param featureIndex The index of the feature in each Obseration to
                            consider a threshold for.
     * @return The best threshold for the specified feature.
     */
    public double findBestThreshold(ArrayList<Observation> observations, 
                             int featureIndex)
    {
        return 0.0;
    }

    /**
     * Evaluates the effectiveness of split1 and split2 for partioning 
     * observations by label.

     * @param split1 A list of Observations.
     * @param split2 A list of Observations.
     * @return A score of how good the splits are. Higher is better.
     */
    public double evaluate(ArrayList<Observation> split1, 
                           ArrayList<Observation> split2){
        return 0.0;
    }

    /**
     * Saves a model of the decision tree (`self.root`) to the given file.
     * This file can be used with the `loadModel()` function.
     * 
     * @param filename The name of the file to save the model to.
     */
    public void saveModel(String filename){

    }

    /**
     * Loads a decision tree model into this instance.
     * 
     * @param filename The name of the file to load the model from.
     */
    public void loadModel(String filename){

    }

    /**
     * Runs the observation through the decision tree and produces a class
     * label prediction.
     * 
     * @param observation The Observation to classify.
     * @return The predicted class label of `observation`.
     */
    public String predict(Observation observation){
        return "";
    }

    public static void main(String[] args) throws IOException {
        final String USAGE = 
            "There are two modes to this: training and prediction.\n\n"+
            "Training:\n"+
            "Usage: DecisionTree -train <training file> <model output file>\n"+
            "where...\n"+
            "   <training file> is a comma separated table of features and a label\n"+
            "                   in the final column; header included\n"+
            "   <model output file> is the name of the file to write the trained model to\n"+
            "\n"+
            "Prediction:\n"+
            "Usage: decision-tree.py -predict <testing file> <model file>\n"+
            "where...\n"+
            "   <testing file> is a comma separated table of features in the same\n"+
            "                  order used during training. Optionally, a label can\n"+
            "                  be in the final column; a header should be present\n"+
            "   <model file> should contain the decision tree model to use for prediction\n"+
            "\n"+
            "The output of the of prediction is the <testing file> data with a new\n"+
            "column: predicted_label";
        

        DecisionTree tree = new DecisionTree();
        String trainingFilename, testingFilename, modelFilename;
        Dataset trainData, testData;

        // Check that enough arguments were specified.
        if(args.length < 3){
            System.err.println("Too few arguments.\n");
            System.err.println(USAGE);
            System.exit(1);
        }

        // Training mode.
        if(args[0].equals("-train")){
            trainingFilename = args[1];
            modelFilename = args[2];

            trainData = tree.parseDataFile(trainingFilename, true);
            tree.train(trainData.observations);
            tree.saveModel(modelFilename);

        // Prediction mode.
        } else if(args[0].equals("-predict")){
            testingFilename = args[1];
            modelFilename = args[2];

            testData = tree.parseDataFile(testingFilename, true);
            tree.loadModel(modelFilename);

            // Print out each testing observation and its predicted label.
            System.out.println(testData.columnNamesAsCSV() +
                ",predicted_label");
            for(Observation obs : testData.observations)
                System.out.println(obs +","+ tree.predict(obs));

        // Unknown mode.
        } else {
            System.err.println("Unrecognized mode: "+ args[0] +"\n");
            System.err.println(USAGE);
            System.exit(1);
        }







    }
}